"""
This module provides functionalities for projecting high-dimensional data (embeddings)
into a lower-dimensional space for visualization, using UMAP (Uniform Manifold Approximation and Projection).
"""

import os
from typing import Tuple, List

import numpy as np
import umap
import pandas as pd
import plotly.graph_objs as go
from tqdm import tqdm

from .constants import VISUALISATION_SETTINGS, PLOT_SIZE

os.environ['TOKENIZERS_PARALLELISM'] = 'false'

def set_up_umap(embeddings: np.ndarray) -> umap.UMAP:
    """
    Sets up and fits a UMAP transformer to the given embeddings.

    Args:
        embeddings (np.ndarray): An array of embeddings to fit the UMAP transformer.

    Returns:
        umap.UMAP: A fitted UMAP transformer.
    """
    umap_transform = umap.UMAP().fit(embeddings)
    return umap_transform

def get_projections(embedding: np.ndarray, umap_transform: umap.UMAP) -> Tuple[np.ndarray, np.ndarray]:
    """
    Projects embeddings into a two-dimensional space using the provided UMAP transformer.

    Args:
        embedding (np.ndarray): An array of embeddings to project.
        umap_transform (umap.UMAP): A fitted UMAP transformer.

    Returns:
        Tuple[np.ndarray, np.ndarray]: X and Y coordinates of the projected embeddings.
    """
    projections = _project_embeddings(embedding, umap_transform)
    x = projections[:, 0]
    y = projections[:, 1]
    return x, y

def _project_embeddings(embeddings: np.ndarray, umap_transform: umap.UMAP) -> np.ndarray:
    """
    Helper function to project embeddings using a UMAP transformer.

    Args:
        embeddings (np.ndarray): An array of embeddings to project.
        umap_transform (umap.UMAP): A fitted UMAP transformer.

    Returns:
        np.ndarray: An array of projected embeddings.
    """
    umap_embeddings = np.empty((len(embeddings), 2))
    for i, embedding in tqdm(enumerate(embeddings), total=len(embeddings)):
        umap_embeddings[i] = umap_transform.transform([embedding])
    return umap_embeddings

def prepare_projections_df(document_ids: List[str], document_projections: Tuple[np.ndarray, np.ndarray], document_text: List[str]) -> pd.DataFrame:
    """
    Prepares a DataFrame for visualization from document IDs, projections, and texts.

    Args:
        document_ids (List[str]): List of document IDs.
        document_projections (Tuple[np.ndarray, np.ndarray]): Tuple of X and Y coordinates of document projections.
        document_text (List[str]): List of document texts.

    Returns:
        pd.DataFrame: DataFrame containing the information for visualization.
    """
    df = pd.DataFrame({"id": document_ids,
                       "x": document_projections[0],
                       "y": document_projections[1]})
    
    df['document'] = document_text
    df['document_cleaned'] = df.document.str.wrap(80).apply(lambda x: x.replace('\n', '<br>'))
    df['size'] = PLOT_SIZE
    df['category'] = "Chunks"
    
    return df

def plot_embeddings(df: pd.DataFrame) -> go.Figure:
    """
    Creates a Plotly figure to visualize the embeddings.

    Args:
        df (pd.DataFrame): DataFrame containing the data to visualize.

    Returns:
        go.Figure: A Plotly figure object for visualization.
    """
    fig = go.Figure()

    for category in df['category'].unique():
        category_df = df[df['category'] == category]
        settings = VISUALISATION_SETTINGS.get(category, {'color': 'grey', 'opacity': 1, 'symbol': 'circle', 'size': 10})
        fig.add_trace(go.Scatter(
            x=category_df['x'],
            y=category_df['y'],
            mode='markers',
            name=category,
            marker=dict(
                color=settings['color'],
                opacity=settings['opacity'],
                symbol=settings['symbol'],
                size=settings['size'],
                line_width=0
            ),
            hoverinfo='text',
            text=category_df['document_cleaned']
        ))

    fig.update_layout(
        height=500,
        legend=dict(
            y=100,
            x=0.5,
            xanchor='center',
            yanchor='top',
            orientation='h'
        )
    )
                
    return fig



from setuptools import setup, find_packages

setup(
    name='ragxplorer',
    version='0.1.6',
    author='Gabriel Chua',
    author_email='cyzgab@gmail.com',
    description='A open-source tool to to visualise your RAG documents 🔮.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/gabrielchua/ragxplorer',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'umap-learn',
        'sentence-transformers',
        'plotly',
        'tqdm',
        'PyPDF2',
        'langchain',
        'chromadb',
        'openai',
        'pydantic'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)



"""
Ragxplorer.py
"""
import os
from typing import (
    Optional,
    Any
    )

from pydantic import BaseModel, Field
import pandas as pd

from chromadb.utils.embedding_functions import (
    SentenceTransformerEmbeddingFunction,
    OpenAIEmbeddingFunction,
    HuggingFaceEmbeddingFunction
    )

import plotly.graph_objs as go

from .rag import (
    build_vector_database,
    get_doc_embeddings,
    get_docs,
    query_chroma
    )

from .projections import (
    set_up_umap,
    get_projections,
    prepare_projections_df,
    plot_embeddings
    )

from .query_expansion import (
    generate_hypothetical_ans,
    generate_sub_qn
    )

from .constants import OPENAI_EMBEDDING_MODELS


class _Documents(BaseModel):
    text: Optional[Any] = None
    ids: Optional[Any] = None
    embeddings: Optional[Any] = None
    projections: Optional[Any] = None

class _Query(BaseModel):
    original_query: Optional[Any] = None
    original_query_projection: Optional[Any] = None
    actual_search_queries: Optional[Any] = None
    retrieved_docs: Optional[Any] = None

class _VizData(BaseModel):
    base_df: Optional[Any] = None
    query_df: Optional[Any] = None
    visualisation_df: Optional[Any] = None

class RAGxplorer(BaseModel):
    """
    RAGxplorer class for managing the RAG exploration process.
    """
    embedding_model: Optional[str] = Field(default="all-MiniLM-L6-v2")
    _chosen_embedding_model: Optional[Any] = None
    _vectordb: Optional[Any] = None
    _documents: _Documents = _Documents()
    _projector: Optional[Any] = None
    _query: _Query = _Query()
    _VizData: _VizData = _VizData()

    def __init__(self, **data):
        super().__init__(**data)
        self._set_embedding_model()

    def _set_embedding_model(self):
        """ Sets the embedding model """
        if self.embedding_model == 'all-MiniLM-L6-v2':
            self._chosen_embedding_model = SentenceTransformerEmbeddingFunction()

        elif self.embedding_model in OPENAI_EMBEDDING_MODELS:
            if "OPENAI_API_KEY" not in os.environ:
                raise OSError("OPENAI_API_KEY is not set")
            self._chosen_embedding_model = OpenAIEmbeddingFunction(api_key = os.getenv("OPENAI_API_KEY"), 
                                                                   model_name = self.embedding_model)
        else:
            try:
                if "HF_API_KEY" not in os.environ:
                    raise OSError("HF_API_KEY is not set")
                self._chosen_embedding_model = HuggingFaceEmbeddingFunction(api_key = os.getenv("HF_API_KEY"),
                                                                            model_name = self.embedding_model)
            except Exception as exc:
                raise ValueError("Invalid embedding model. Please use all-MiniLM-L6-v2, or a valid OpenAI or HuggingFace embedding model.") from exc

    def load_pdf(self, document_path: str, chunk_size: int = 1000, chunk_overlap: int = 0, verbose: bool = False):
        """
        Load data from a PDF file and prepare it for exploration.
        
        Args:
            document: Path to the PDF document to load.
            chunk_size: Size of the chunks to split the document into.
            chunk_overlap: Number of tokens to overlap between chunks.
        """
        if verbose:
            print(" ~ Building the vector database...")
        self._vectordb = build_vector_database(document_path, chunk_size, chunk_overlap, self._chosen_embedding_model)
        if verbose:
            print("Completed Building Vector Database ✓")
        self._documents.embeddings = get_doc_embeddings(self._vectordb)
        self._documents.text = get_docs(self._vectordb)
        self._documents.ids = self._vectordb.get()['ids']
        if verbose:
            print(" ~ Reducing the dimensionality of embeddings...")
        self._projector = set_up_umap(embeddings=self._documents.embeddings)
        self._documents.projections = get_projections(embedding=self._documents.embeddings,
                                                      umap_transform=self._projector)
        self._VizData.base_df = prepare_projections_df(document_ids=self._documents.ids,
                                                                document_projections=self._documents.projections,
                                                                document_text=self._documents.text)
        if verbose:
            print("Completed reducing dimensionality of embeddings ✓")

    def visualize_query(self, query: str, retrieval_method: str="naive", top_k:int=5, query_shape_size:int=5) -> go.Figure:
        """
        Visualize the query results in a 2D projection using Plotly.

        Args:
            query (str): The query string to visualize.
            retrieval_method (str): The method used for document retrieval. Defaults to 'naive'.
            top_k (int): The number of top documents to retrieve.
            query_shape_size (int): The size of the shape to represent the query in the plot.

        Returns:
            go.Figure: A Plotly figure object representing the visualization.

        Raises:
            RuntimeError: If the document has not been loaded before visualization.
        """
        if self._vectordb is None or self._VizData.base_df is None:
            raise RuntimeError("Please load the pdf first.")

        self._query.original_query = query

        if (self.embedding_model == "all-MiniLM-L6-v2") or (self.embedding_model in OPENAI_EMBEDDING_MODELS):
            self._query.original_query_projection = get_projections(embedding=self._chosen_embedding_model(self._query.original_query),
                                                                umap_transform=self._projector)
        else:
            self._query.original_query_projection = get_projections(embedding=[self._chosen_embedding_model(self._query.original_query)],
                                                                    umap_transform=self._projector)

        self._VizData.query_df = pd.DataFrame({"x": [self._query.original_query_projection[0][0]],
                                      "y": [self._query.original_query_projection[1][0]],
                                      "document_cleaned": query,
                                      "category": "Original Query",
                                      "size": query_shape_size})

        if retrieval_method == "naive":
            self._query.actual_search_queries = self._query.original_query

        elif retrieval_method == "HyDE":
            if "OPENAI_API_KEY" not in os.environ:
                raise OSError("OPENAI_API_KEY is not set")
            self._query.actual_search_queries = generate_hypothetical_ans(query=self._query.original_query)

        elif retrieval_method == "HyDE":
            if "OPENAI_API_KEY" not in os.environ:
                raise OSError("OPENAI_API_KEY is not set")
            self._query.actual_search_queries = generate_sub_qn(query=self._query.original_query)

        self._query.retrieved_docs = query_chroma(chroma_collection=self._vectordb,
                                            query=self._query.actual_search_queries,
                                            top_k=top_k)

        self._VizData.base_df.loc[self._VizData.base_df['id'].isin(self._query.retrieved_docs), "category"] = "Retrieved"
            
        self._VizData.visualisation_df = pd.concat([self._VizData.base_df, self._VizData.query_df], axis = 0)

        return plot_embeddings(self._VizData.visualisation_df)
